!*****************************************************************************!
!* 
!*      $RCSfile: IniSod.F90,v $
!*
!*      $Revision: 1.5 $
!*
!*      $Date: 2004/03/08 18:00:10 $
!*
!*      $Author: roller $
!*
!*****************************************************************************!

MODULE IniEuler2D_mod
  !---------------------------------------------------------------------------!
  IMPLICIT NONE
  PRIVATE
  !---------------------------------------------------------------------------!
  INTERFACE IniEuler2D
     MODULE PROCEDURE IniEuler2D
  END INTERFACE
  !---------------------------------------------------------------------------!
  PUBLIC  :: IniEuler2D
  !---------------------------------------------------------------------------!

CONTAINS

  SUBROUTINE IniEuler2D(pvar,cvar,iter,dt,time,printtime,Const,Mesh,FileIO,MPI)
    USE Types
    !---------------------------------------------------------------------------!
    IMPLICIT NONE
    !---------------------------------------------------------------------------!
#ifdef parallel
    include 'mpif.h'
#endif
    !---------------------------------------------------------------------------!
    ! Argument list declaration
    TYPE(tConstants)        :: Const
    TYPE(tMesh)             :: Mesh
    TYPE(tFileIO)           :: FileIO
    TYPE(tMPI)              :: MPI
    REAL, POINTER           :: pvar(:,:,:)
    REAL, POINTER           :: cvar(:,:,:)
    INTEGER                 :: iter
    REAL                    :: dt
    REAL                    :: time
    REAL                    :: printtime
    ! Local variable declaration
    INTEGER                 :: allocStat
    !---------------------------------------------------------------------------!

    ! (0): MPI Initialization                                                   !
    ! --------------------------------------------------------------------------!
    
#ifdef parallel
    call mpi_init(MPI%ierror)
    call mpi_comm_rank(mpi_comm_world,MPI%myrank,MPI%ierror)
    call mpi_comm_size(mpi_comm_world,MPI%mpisize,MPI%ierror)
#else
    MPI%myrank = 0
#endif

    if (MPI%myrank ==0) then
       WRITE(*,*) '============================================================'
       WRITE(*,*) 'Solution of the 2D Euler equations with Finite Volume Method'
       WRITE(*,*) '============================================================'
       WRITE(*,*)
       WRITE(*,*) 'Initializing program....'
    endif

    ! (1): General constants                                                    !
    ! --------------------------------------------------------------------------!

    Const%Pi        = 3.1415926
    Const%gamma     = 1.4
    Const%nvar      = 4
    FileIO%uOutFile = 777

    ! (2): Initialization of tMesh                                              !
    ! --------------------------------------------------------------------------!

    Mesh%IMAX = 100                  ! Number of cells in x-direction           !
    Mesh%JMAX = 1                    ! Number of cells in y-direction           !
    Mesh%x0   = 0.                   ! x Koordinatenursprung                    !
    Mesh%y0   = 0.                   ! y Koordinatenursprung                    !
    Mesh%xlen = 1.                   ! Lnge des Rechengebiets in x-Richtung    !
    Mesh%ylen = 0.01                 ! Lnge des Rechengebiets in y-Richtung    !
    
    ! (3): Initialization of tConstants                                         !
    ! --------------------------------------------------------------------------!

    Const%CFL        = 0.9          ! CFL number                                !
    Const%MaxIter    = 1500         ! Maximum iteration number                  !
    Const%Stoptime   = 0.25         ! Final simulation time                     !
    Const%icType     = 5            ! Sod problem                               !
    Const%TimeOrder  = 1            ! Second Order time discretisation          !
    Const%SpaceOrder = 1            ! Second Order spatial discretisation       !
    Const%iFlux      = 1            ! Flux function: Godunov                    !
    ! Implemented Boundary Conditions:
    Const%BoundCond = 1   ! Extrapolation 1st order at all boundaries
    ! Const%BoundCond = 2 ! Extrapolation 1st order in x, periodic in y-direction
    ! Const%BoundCond = 3 ! Periodic in x-direction, extrapolation 1st order in y
    ! Const%BoundCond = 4 ! Periodic in x- and y-direction

    ! Constants concerning initial condition
    Const%amplitude  = 1.
    Const%hwidth     = 6.

    ! (4): Initialization of tFileIO                                            !
    ! --------------------------------------------------------------------------!

    FileIO%strOutFile = 'Sod'       ! Filename for data output                  !
    FileIO%iOutput    = 0.25        ! Timeinterval for data output              !
    FileIO%iVisuProg  = 2           ! (1) = Tecplot, (2) = IBM Open DX

    ! (5): Initialization of other variables                                    !
    ! --------------------------------------------------------------------------!

    iter = 0                        ! Set iteration to zero (initial condition) !
    dt   = 0.                       ! Set time step to zero       (ini. cond.)  !
    time = 0.                       ! Set simulation time to zero (ini. cond.)  !
    printtime = 0.                  ! Set printtime to zero       (ini. cond.)  !

    ! (6): MPI Vorbereitungen                                                   !
    ! --------------------------------------------------------------------------!
    
    MPI%IMAX = Mesh%IMAX            ! Gesamt-Anzahl aller Punkte sichern        !
    MPI%JMAX = Mesh%JMAX

                                    ! Definiere Offset auch fr seriellen Fall  !
                                    ! um if-Abfragen zu vermeiden.              !
    MPI%xOffset = 0
    MPI%yOffset = 0

#ifdef parallel
                                    ! Erzeugen eines kartesischen Gitters von   !
                                    ! Prozessoren                               !
                                    !                                           !
                                    ! 1. Schritt: Wieviele Knoten je Richtung   !
    call mpi_dims_create(MPI%mpisize,2,MPI%dims,MPI%ierror)
    if (MPI%myrank == 0)  then
       write(*,*)
       write(*,*) 'MPI-Info: '
       write(*,'(a,i3,a,i3,a,i4,a)') '   ',MPI%dims(1),' x',MPI%dims(2),   &
            ' = ',MPI%mpisize,' nodes are used. '
       write(*,*)
    endif
                                    ! 2. Schritt: Erzeuge neuen Kommunikator    !

    select case (Const%BoundCond)
    case (1)
       MPI%periods(:) = .false.     ! Extrapolation in x,y
    case (2)
       MPI%periods(1) = .false.     ! Extrapolation in x 
       MPI%periods(2) = .true.      ! Periodisch in y
    case (3)
       MPI%periods(1) = .true.      ! Periodisch in x
       MPI%periods(2) = .false.     ! Extrapolation in y
    case (4)
       MPI%periods(:) = .true.      ! Periodisch in x,y
    case default
       write(*,*) 'No other boundary condition known'
       STOP
    end select

    call mpi_cart_create    &
         (mpi_comm_world,2,MPI%dims,MPI%periods,.true.,MPI%comm_cart,MPI%ierror)
                                    ! 3. Schritt: Welcher Prozess bin ich?      !
                                    !          a: von allen (1dim)              !
                                    !             z.B. Prozess 5 von 8          !
    call mpi_comm_rank(MPI%comm_cart,MPI%myrank,MPI%ierror)
                                    !          b: in 2D-Notation                !
                                    !             0=(0,0) 2=(1,0) 4=(2,0) 6=(3,0)
                                    !             1=(0,1) 3=(1,1) 5=(2,1) 7=(3,1)
                                    ! Achtung: gezhlt wird wie in C,           !
                                    !          d.h. 0..n-1 und spaltenweise     !
                                    !          im Gegensatz zu Fortran          !
                                    !          d.h. 1..n und zeilenweise        !

    call mpi_cart_coords(MPI%comm_cart,MPI%myrank,2,MPI%coords,MPI%ierror)

                                    ! Verteilen der Zellen auf die Prozessoren  !
    MPI%xNumber = int(real(MPI%IMAX) / MPI%dims(1)+0.5)
    MPI%yNumber = int(real(MPI%JMAX) / MPI%dims(2)+0.5)

    if (MPI%coords(1) .ne. (MPI%dims(1)-1)) then
       Mesh%IMAX = MPI%xNumber
    else                            ! der letzte kriegt mehr oder weniger Zellen
       Mesh%IMAX = MPI%IMAX - MPI%xNumber*(MPI%dims(1)-1)
    endif
    if (MPI%coords(2) .ne. (MPI%dims(2)-1)) then
       Mesh%JMAX = MPI%yNumber
    else                            ! der letzte kriegt mehr oder weniger Zellen
       Mesh%JMAX = MPI%JMAX - MPI%yNumber*(MPI%dims(2)-1)
    endif

                                    ! wieviel Zellen liegen vor meinem Gebiet
    MPI%xOffset = MPI%xNumber*MPI%coords(1)
    MPI%yOffset = MPI%yNumber*MPI%coords(2)
#endif

    ! (7): Memory allocation                                                    !
    ! --------------------------------------------------------------------------!
    
    ALLOCATE(pvar(0:Mesh%IMAX+1,0:Mesh%JMAX+1,Const%nvar),      &
             cvar(0:Mesh%IMAX+1,0:Mesh%JMAX+1,Const%nvar),      &
             Mesh%Bary(0:Mesh%IMAX+1,0:Mesh%JMAX+1,2),          &
             Mesh%Surface(0:Mesh%IMAX+1,0:Mesh%JMAX+1),         &
             STAT = allocStat )

    IF(allocStat.NE.0) THEN
       PRINT *, 'ERROR IniEuler2D: Could not allocate variables!'
       STOP
    ENDIF


    
  END SUBROUTINE IniEuler2D


END MODULE IniEuler2D_mod
